from typing import List, Dict
import glob
from tqdm import tqdm

from blueness import module
from blue_objects import file, objects
from blue_objects.graphics.gif import generate_animated_gif
from blue_objects.metadata import post_to_object

from blue_geo import NAME
from blue_geo.catalog.generic.generic.scope import raster_suffix
from blue_geo.watch.workflow.common import load_watch
from blue_geo.logger import logger


NAME = module.name(__file__, NAME)


def reduce_function(
    query_object_name: str,
    suffix: str,
    object_name: str,
    content_threshold: float = 0.5,
    list_of_suffix: List[str] = raster_suffix,
) -> bool:
    success, target, list_of_files = load_watch(
        object_name,
        list_of_suffix=list_of_suffix,
    )
    if not success:
        return success

    logger.info(
        "{}.reduce {}/{} @ {} -{} file(s)-> {}".format(
            NAME,
            query_object_name,
            suffix,
            target.one_liner,
            len(list_of_files),
            object_name,
        )
    )

    logger.info("loading metadata ...")
    frame_metadata: Dict[str, Dict] = {}
    bad_metadata: List[str] = []
    for filename in tqdm(
        glob.glob(
            objects.path_of(
                "metadata-*.yaml",
                object_name,
            )
        )
    ):
        success, metadata_ = file.load_yaml(filename)
        if not success:
            bad_metadata.append(filename)
            continue

        if not "filename" in metadata_.get("map", {}):
            logger.info('no "filename" in metadata["map"]: {}.'.format(filename))
            bad_metadata.append(file.name_and_extension(filename))
            continue

        frame_metadata[metadata_["map"]["filename"]] = metadata_
    if bad_metadata:
        logger.info("bad metadata: {}.".format(", ".join(bad_metadata)))

    bad_frames: List[str] = []
    low_content_frames: List[str] = []
    list_of_frames: List[str] = []
    for _, filename in enumerate(list_of_files):
        frame_metadata.setdefault(file.name_and_extension(filename), {})

        frame_content_ratio = (
            frame_metadata[file.name_and_extension(filename)]
            .get("map", {})
            .get("content_ratio", -1)
        )
        frame_is_usable = (
            frame_metadata[file.name_and_extension(filename)]
            .get("map", {})
            .get("usable", False)
        )
        if not frame_is_usable:
            bad_frames.append(file.name_and_extension(filename))
            continue

        frame_filename = file.add_extension(filename, "png")

        frame_has_content = bool(frame_content_ratio >= content_threshold)
        logger.info(
            "{} / {}: content={:.03f} {} {:.03f}".format(
                "✅" if frame_has_content else "🛑",
                file.name_and_extension(filename),
                frame_content_ratio,
                ">=" if frame_has_content else "<",
                content_threshold,
            )
        )

        frame_metadata[file.name_and_extension(filename)].update(
            {
                "has_content": bool(frame_has_content),
            }
        )

        if not frame_has_content:
            low_content_frames.append(file.name_and_extension(filename))
            continue

        list_of_frames.append(frame_filename)
    if bad_frames:
        logger.error("{} bad frame(s).".format(len(bad_frames)))

    if not post_to_object(
        object_name,
        "reduce",
        {
            "algo": "modality",
            "bad_frames": bad_frames,
            "bad_metadata": bad_metadata,
            "frames": frame_metadata,
            "low_content_frames": low_content_frames,
            "list_of_files": [
                file.name_and_extension(filename) for filename in list_of_files
            ],
            "target": target.__dict__,
        },
    ):
        return False

    return all(
        generate_animated_gif(
            list_of_frames,
            objects.path_of(
                "{}{}.gif".format(
                    object_name,
                    f"-{scale}X" if scale != 1 else "",
                ),
                object_name,
            ),
            frame_duration=1000,
            scale=scale,
        )
        for scale in [1, 2, 4]
    )



from blue_geo.help.catalog import help_functions as help_catalog
from blue_geo.help.datacube import help_functions as help_datacube
from blue_geo.help.gdal import help_functions as help_gdal
from blue_geo.help.QGIS import help_functions as help_QGIS
from blue_geo.help.watch import help_functions as help_watch

help_functions = {
    "catalog": help_catalog,
    "datacube": help_datacube,
    "gdal": help_gdal,
    "QGIS": help_QGIS,
    "watch": help_watch,
}



def generate_seed() -> str:
    list_of_modules = [
        "log",
        "project",
        "layer",
        "application",
        "seed",
        "QGIS",
        "apps/vanwatch",
        "apps/template",
        "apps/ukraine_timemap",
        "main",
        "utils",
    ]

    seed = "; ".join(
        [
            'exec(Path(f\'{os.getenv("HOME")}/git/blue-geo/blue_geo/QGIS/console/'
            + module_name
            + ".py').read_text())"
            for module_name in list_of_modules
        ]
    )

    return seed



from typing import Dict
from tqdm import trange
import numpy as np
import matplotlib.pyplot as plt
import cv2

from blueness import module
from blue_objects import file, objects
from blue_objects.metadata import post_to_object
from blue_objects.graphics.signature import justify_text, add_signature

from blue_geo import NAME
from blue_geo.host import signature
from blue_geo.catalog.generic import GenericDatacube
from blue_geo.watch.workflow.common import load_watch
from blue_geo.logger import logger


NAME = module.name(__file__, NAME)


def map_function(
    query_object_name: str,
    suffix: str,
    offset: str,
    depth: int,
    diff_range: float = 100,
    line_width: int = 80,
    colorbar_width: int = 20,
) -> bool:
    diff_filename: str = ""
    if depth < 2:
        logger.error(f"depth={depth} < 2!")
        return False

    offset_int = int(offset)

    object_name = f"{query_object_name}-{suffix}-{offset}"

    success, target, _ = load_watch(object_name, log=False)
    if not success:
        return success

    logger.info(
        "{}.map: {} #{}+D={} -> {}".format(
            NAME,
            target.one_liner,
            offset,
            depth,
            object_name,
        )
    )

    acquisition_metadata: Dict[str, Dict] = {}
    baseline_filename: str = ""
    target_filename: str = ""
    for index in trange(depth):
        logger.info(f"processing metadata: index={index} ...")

        index_object_name = "{}-D-{:03d}".format(
            object_name,
            offset_int + index,
        )
        success, acquisition_metadata[index] = file.load_yaml(
            objects.path_of(
                filename="metadata.yaml",
                object_name=index_object_name,
            ),
            ignore_error=True,
        )
        if not success:
            break

        if index in [0, 1]:
            filename = objects.path_of(
                filename=acquisition_metadata[index].get("map", {}).get("filename", ""),
                object_name=index_object_name,
            )
            if not file.exists(filename):
                logger.warning(f"{filename} not found!")
                success = False
                break

            if index == 0:
                baseline_filename = filename
            else:
                target_filename = filename

    if success:
        success, baseline_image, _ = GenericDatacube.load_modality(
            filename=baseline_filename,
            modality="rgb",
            log=True,
        )

    if success:
        success, target_image, _ = GenericDatacube.load_modality(
            filename=target_filename,
            modality="rgb",
            log=True,
        )

    if success:
        diff_image = np.squeeze(target_image[:, :, 0] - baseline_image[:, :, 0]).astype(
            np.float32
        )
        diff_image[diff_image < -diff_range] = -diff_range
        diff_image[diff_image > diff_range] = diff_range

        plt.figure(figsize=(10, 6))
        plt.hist(
            diff_image.ravel(),
            bins=256,
            range=(-diff_range, diff_range),
        )
        plt.title(
            justify_text(
                " | ".join(
                    [
                        "diff histogram",
                        query_object_name,
                        f"/{suffix}",
                        f"@{offset}+{depth}",
                        f"+-{diff_range:.2f}",
                        file.name_and_extension(baseline_filename),
                        file.name_and_extension(target_filename),
                    ]
                ),
                line_width=line_width,
                return_str=True,
            )
        )
        plt.xlabel(
            justify_text(
                " | ".join(["DN diff"] + signature()),
                line_width=line_width,
                return_str=True,
            )
        )
        plt.ylabel("frequency")
        plt.grid(True)
        success = file.save_fig(
            objects.path_of(
                "{}-diff-histogram.png".format(file.name(target_filename)),
                object_name,
            ),
            log=True,
        )

    if success:
        colored_diff = cv2.applyColorMap(
            ((diff_image / diff_range + 1) / 2 * 255).astype(np.uint8), cv2.COLORMAP_JET
        )

        gradient = (
            255
            * np.linspace(0, 1, colored_diff.shape[0]).reshape(-1, 1)
            * np.ones((1, colorbar_width))
        ).astype(np.uint8)
        colorbar = cv2.applyColorMap(gradient, cv2.COLORMAP_JET)
        concatenated_image = np.hstack(
            (
                colored_diff,
                np.zeros(
                    (colored_diff.shape[0], colorbar_width // 2, 3),
                    dtype=np.uint8,
                ),
                colorbar,
            )
        )
        colored_diff_signed = add_signature(
            concatenated_image,
            header=[
                " | ".join(
                    objects.signature(
                        " | ".join(
                            [
                                suffix,
                                f"@{offset}+{depth}",
                                f"+-{diff_range:.2f}",
                                file.name_and_extension(baseline_filename),
                                file.name_and_extension(target_filename),
                            ]
                        ),
                        query_object_name,
                    )
                ),
            ],
            footer=[
                target.one_liner,
                " | ".join(signature()),
            ],
            word_wrap=True,
        )
        diff_filename = objects.path_of(
            "{}-diff.png".format(file.name(target_filename)),
            object_name,
        )
        success = file.save_image(
            diff_filename,
            colored_diff_signed,
            log=True,
        )

    if not success:
        logger.warning("not usable.")

    return post_to_object(
        object_name,
        "map",
        {
            "algo": "diff",
            "inputs": {
                "offset": offset,
                "depth": depth,
                "acquisitions": acquisition_metadata,
            },
            "filename": file.name_and_extension(diff_filename),
            "target": target.__dict__,
            "usable": success,
        },
    )



import folium
import geopandas as gpd
from IPython.display import display


def show_on_map(
    gdf: gpd.GeoDataFrame,
    zoom_start: int = 12,
):
    if gdf.empty:
        return

    map = folium.Map(
        location=gdf.unary_union.centroid.coords[0][::-1],
        zoom_start=zoom_start,
    )
    folium.GeoJson(gdf).add_to(map)

    display(map)



import numpy as np
import cv2
import math

from blueness import module
from blue_options import string
from blue_objects import file, objects
from blue_objects.metadata import post_to_object, get_from_object
from blue_objects.graphics.signature import add_signature

from blue_geo import NAME
from blue_geo.catalog.functions import get_datacube_class
from blue_geo.host import signature
from blue_geo.watch.workflow.common import load_watch
from blue_geo.logger import logger


NAME = module.name(__file__, NAME)


def map_function(
    query_object_name: str,
    suffix: str,
    offset: str,
    modality: str,
    min_width: int = 1200,
) -> bool:
    offset_int = int(offset)

    object_name = f"{query_object_name}-{suffix}-{offset}"

    list_of_datacube_id = get_from_object(
        query_object_name,
        "datacube_id",
        [],
    )
    if len(list_of_datacube_id) < offset_int + 1:
        logger.warning(f"offset={offset}: datacube-id not found.")
        return True
    datacube_id = list_of_datacube_id[offset_int]

    success, target, list_of_files = load_watch(object_name)
    if not success or not list_of_files:
        return success
    filename = list_of_files[0]

    logger.info(
        "{}.map: {} #{} @ {} -{}-> {}".format(
            NAME,
            target.one_liner,
            offset,
            datacube_id,
            modality,
            object_name,
        )
    )

    datacube_class = get_datacube_class(datacube_id)
    success, frame, _ = datacube_class.load_modality(
        filename,
        modality=modality,
        log=True,
    )

    if min_width != -1 and frame.shape[1] < min_width and frame.shape[1] > 0:
        scale = int(math.ceil(min_width / frame.shape[1]))

        logger.info(
            "scaling {} X {}".format(
                string.pretty_shape_of_matrix(frame),
                scale,
            )
        )

        frame = cv2.resize(
            frame,
            (
                scale * frame.shape[1],
                scale * frame.shape[0],
            ),
            interpolation=cv2.INTER_NEAREST_EXACT,
        )

    frame_fp32 = frame.astype(np.float32).flatten() / 255
    frame_fp32 = frame_fp32[frame_fp32 > 0.0]
    frame_fp32 = frame_fp32[frame_fp32 < 1.0]
    content_ratio = len(frame_fp32) / frame.size

    frame = add_signature(
        frame,
        header=[
            " | ".join(
                objects.signature(
                    "{} / {}".format(
                        datacube_id,
                        file.name_and_extension(filename),
                    ),
                    object_name,
                )
                + [
                    "{:05.1f}%".format(content_ratio * 100.0),
                    f"#{offset}",
                ]
            ),
        ],
        footer=[
            target.one_liner,
            " | ".join(signature()),
        ],
        word_wrap=True,
    )

    frame_filename = file.add_extension(filename, "png")

    success = file.save_image(frame_filename, frame, log=True)

    return post_to_object(
        object_name,
        "map",
        {
            "algo": "modality",
            "inputs": {
                "offset": offset,
                "modality": modality,
            },
            "content_ratio": content_ratio,
            "datacube_id": datacube_id,
            "filename": file.name_and_extension(filename),
            "target": target.__dict__,
            "usable": success,
        },
    )



from typing import Tuple, Any, Dict, List
from datetime import datetime, timedelta
from tqdm import tqdm

from blue_objects.metadata import post_to_object

from blue_geo.catalog.generic.generic.classes import GenericDatacube
from blue_geo.catalog.generic.generic.scope import DatacubeScope
from blue_geo.catalog.generic.stac.classes import STACCatalog
from blue_geo.logger import logger


class STACDatacube(GenericDatacube):
    collection = "unknown"

    QGIS_template = "unknown-template"

    metadata: Any = {}

    query_args = {
        "bbox": {
            "type": str,
            "default": "",
            "help": "<-122.88,51.73,-122.68,51.93>",
        },
        "datetime": {
            "default": "{}/{}".format(
                (datetime.now() - timedelta(days=30)).strftime("%Y-%m-%d"),
                datetime.now().strftime("%Y-%m-%d"),
            ),
            "help": "<2024-07-30/2024-08-09>, more: https://documentation.dataspace.copernicus.eu/APIs/STAC.html#search-items-by-datetime",
        },
        "lat": {
            "type": float,
            "default": 51.83,
            "help": "<51.83>",
        },
        "lon": {
            "type": float,
            "default": -122.78,
            "help": "<-122.78>",
        },
        "count": {
            "type": int,
            "default": -1,
            "help": "<10>, -1: all",
        },
        "radius": {
            "type": float,
            "default": 0.1,
            "help": "<0.1>",
        },
        "keyword": {
            "default": "",
            "help": "<keyword>",
        },
    }

    def ingest(
        self,
        dryrun: bool = False,
        overwrite: bool = False,
        scope: str = "metadata",
        verbose: bool = True,
    ) -> Tuple[bool, Any]:
        success, output = super().ingest(dryrun, overwrite, scope)
        if not success:
            return success, output

        list_of_files = self.list_of_files(DatacubeScope(scope))
        logger.info(
            "ingesting {} file(s){}...".format(
                len(list_of_files),
                " in dryrun mode " if dryrun else "",
            )
        )

        error_count = 0
        for filename in tqdm(list_of_files):
            if dryrun:
                continue

            if not self.ingest_filename(
                filename,
                overwrite=overwrite,
                verbose=verbose,
            ):
                error_count += 1

        if error_count:
            logger.error(f"{error_count} error(s).")

        return error_count == 0, output

    @classmethod
    def parse_datacube_id(cls, datacube_id: str) -> Tuple[
        bool,
        Dict[str, Any],
    ]:
        success, _ = super().parse_datacube_id(datacube_id)
        if not success:
            return False, {}

        # datacube-copernicus-sentinel_1-<item_id>
        segments = datacube_id.split("-")
        if len(segments) < 4:
            return False, {}

        if segments[2] != cls.name:
            return False, {}

        return True, {
            "item_id": datacube_id.split("-", 4)[3],
        }

    @classmethod
    def query(
        cls,
        object_name: str,
        datetime: str,
        bbox: List[float],
        count: int,
        keyword: str = "",
        verbose: bool = False,
    ) -> bool:
        assert isinstance(cls.catalog, STACCatalog)

        logger.info(
            "🔎 {}.query -{}> {}".format(
                cls.__name__,
                f"{keyword}-" if keyword else "",
                object_name,
            )
        )

        success, client = cls.catalog.get_client()
        if not success:
            return success

        search_parameters = {
            "collections": [cls.collection],
            "bbox": bbox,
            "datetime": datetime,
        }
        for param, value in search_parameters.items():
            logger.info(f"🔎 {param}: {value}")

        try:
            search = client.search(**search_parameters)
        except Exception as e:
            logger.error(e)
            return False

        items = list(search.item_collection())
        if verbose:
            for item in items:
                logger.info(
                    "🧊 {}: {} @ {}".format(
                        item.id, item.datetime, ", ".join(list(item.assets.keys()))
                    )
                )

        list_of_datacube_ids: List[str] = [
            "datacube-{}-{}-{}".format(
                cls.catalog.name,
                cls.name,
                item.id.replace(".SAFE", "-SAFE"),
            )
            for item in items
        ][::-1]

        if keyword:
            list_of_datacube_ids = [
                datacube_id
                for datacube_id in list_of_datacube_ids
                if keyword in datacube_id
            ]

        if count != -1:
            list_of_datacube_ids = list_of_datacube_ids[:count]

        logger.info(f"{len(list_of_datacube_ids)} datacubes(s) found.")
        for index, datacube_id in enumerate(list_of_datacube_ids):
            logger.info(f"🧊 {index+1:02}: {datacube_id}")

        return post_to_object(
            object_name,
            "datacube_id",
            list_of_datacube_ids,
        )

    @property
    def raw(self) -> str:
        return "{} @ {}".format(
            super().raw,
            self.metadata["Item"].to_dict(),
        )

    def update_metadata(self, verbose: bool = False) -> bool:
        assert isinstance(self.catalog, STACCatalog)

        if not super().update_metadata(verbose):
            return False

        success, client = self.catalog.get_client()
        if not success:
            return success

        search_parameters = {
            "ids": [self.raw_datacube_id()],
            "collections": [self.collection],
        }

        if verbose:
            for param, value in search_parameters.items():
                logger.info(f"{param}: {value}")

        search = client.search(**search_parameters)

        items = list(search.item_collection())
        if not items:
            return False
        if len(items) != 1 and verbose:
            logger.waring(
                "{}.update_metadata({}): {} instance(s) found.".format(
                    self.__class__.__name__,
                    self.datacube_id,
                    len(items),
                )
            )
        item = items[0]
        if verbose:
            logger.info(
                "🧊 {}: {} @ {}".format(
                    item.id, item.datetime, ", ".join(list(item.assets.keys()))
                )
            )
        self.metadata["Item"] = item

        self.metadata["files"] = self.list_of_files()

        if verbose:
            logger.info(f'{len(self.metadata["files"])} file(s).')
            for index, filename in enumerate(self.metadata["files"]):
                logger.info(f"#{index+1}: {filename}")

        return True



from typing import List
import os

from blue_options.terminal import show_usage


def help_expressions_pull(
    tokens: List[str],
    mono: bool,
) -> str:
    return show_usage(
        "QGIS expressions pull",
        "pull QGIS expressions.",
        mono=mono,
    )


def help_expressions_push(
    tokens: List[str],
    mono: bool,
) -> str:
    return show_usage(
        "QGIS expressions push [push]",
        "push QGIS expressions.",
        {
            "QGIS: {}".format(
                os.getenv(
                    "abcli_QGIS_path_expressions",
                    "",
                )
            ): "",
            "github: {}".format(
                os.getenv(
                    "abcli_QGIS_path_expressions_git",
                    "",
                )
            ): "",
        },
        mono=mono,
    )


def help_seed(
    tokens: List[str],
    mono: bool,
) -> str:
    return show_usage(
        "QGIS seed [screen]",
        "seed 🌱 QGIS.",
        mono=mono,
    )


def help_server(
    tokens: List[str],
    mono: bool,
) -> str:
    return show_usage(
        "QGIS serve[r] [start]",
        "start QGIS server.",
        mono=mono,
    )


help_functions = {
    "expressions": {
        "push": help_expressions_push,
        "pull": help_expressions_pull,
    },
    "seed": help_seed,
    "server": help_server,
}



from typing import List

from blue_options.terminal import show_usage

from blue_geo.catalog import get_catalog
from blue_geo.help.datacube import ingest_options
from blue_geo.catalog.functions import get_datacube_class_in_catalog
from blue_geo.catalog.default import as_list_of_args
from blue_geo.logger import logger


def help_browse(
    tokens: List[str],
    mono: bool,
) -> str:
    catalog_name = tokens[0]
    catalog = get_catalog(catalog_name)

    return show_usage(
        [
            "@catalog browse",
            catalog_name,
            "|".join(catalog.url.keys()),
        ],
        f"browse {catalog_name}.",
        mono=mono,
    )


def help_get(
    tokens: List[str],
    mono: bool,
) -> str:
    what = "is_STAC|url:<...>|url_args|list_of_args"
    args = ["[--catalog <catalog>]"]
    return show_usage(
        [
            "@catalog get",
            f"[{what}]",
        ]
        + args,
        "get catalog properties.",
        mono=mono,
    )


def help_list(
    tokens: List[str],
    mono: bool,
) -> str:
    options = "catalogs"
    args = [
        "[--count 1]",
        "[--delim ,]",
        "[--log 0]",
    ]
    usage_1 = show_usage(
        [
            "@catalog list",
            f"[{options}]",
        ]
        + args,
        "list catalogs.",
        mono=mono,
    )

    options = "collections|datacubes==datacube_classes"
    args = ["[--catalog <catalog>]"] + args
    usage_2 = show_usage(
        [
            "@catalog list",
            f"[{options}]",
        ]
        + args,
        f"list {options} in <catalog>.",
        mono=mono,
    )

    return "\n".join([usage_1, usage_2])


def help_query(
    tokens: List[str],
    mono: bool,
) -> str:
    if not tokens:
        logger.error(f"token expected: {tokens}")
        return ""

    catalog_name = tokens[0]

    if catalog_name == "read":
        return help_query_read(tokens[1:], mono=mono)

    if len(tokens) < 2:
        logger.error(f"2 tokens expected: {tokens}")
        return ""

    datacube_class_name = tokens[1]
    datacube_class = get_datacube_class_in_catalog(
        catalog_name,
        datacube_class_name,
    )
    args = as_list_of_args(datacube_class.query_args)
    options = f"dryrun,{datacube_class_name},select,upload"

    return show_usage(
        [
            f"@catalog query {catalog_name}",
            f"[{options}]",
            f"[ingest,{ingest_options(mono)}]",
            "[-|<object-name>]",
        ]
        + args,
        f"{catalog_name}/{datacube_class_name} -query-> <object-name>.",
        {
            "scope: @datacube ingest help.": "",
        },
        mono=mono,
    )


def help_query_read(
    tokens: List[str],
    mono: bool,
) -> str:
    options = "all,download,len"
    args = [
        "[--count <count>]",
        "[--delim <delim>]",
        "[--offset <offset>]",
        "[--prefix <prefix>]",
        "[--suffix <suffix>]",
        "[--contains <contains>]",
        "[--notcontains <not-contains>]",
    ]

    return show_usage(
        [
            "@catalog query read",
            f"[{options}]",
            "[.|<object-name>]",
        ]
        + args,
        "read query results in <object-name>.",
        mono=mono,
    )


help_functions = {
    "browse": help_browse,
    "get": help_get,
    "list": help_list,
    "query": help_query,
}



from typing import List, Tuple, Union
from pystac_client import Client

from blue_geo.catalog.generic.classes import GenericCatalog
from blue_geo.logger import logger


class STACCatalog(GenericCatalog):

    @classmethod
    def get_client(cls) -> Tuple[bool, Union[Client, None]]:
        try:
            # https://pystac-client.readthedocs.io/en/stable/
            client = Client.open(cls.url["api"])
        except Exception as e:
            logger.error(e)
            return False, None

        return True, client

    def get_list_of_collections(self) -> List[str]:
        success, client = self.get_client()
        if not success:
            return []

        return sorted([collection.id for collection in client.get_collections()])
